#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
#  OX for UCS
#  listener module: handling users
#
# Copyright (C) 2004-2017 Univention GmbH <http://www.univention.de/>
# and
#     iKu Systems & Services GmbH & Co. KG (http://www.iku-systems.de/)
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

__package__ = ''  # workaround for PEP 366
import re
import os
import subprocess
from ldap.filter import filter_format

import listener
import univention.uldap
from univention.ox.oxlogger import OXLogger
from univention.ox.listener_tools import ListenerQueue, AsRoot, get_credentials, ListenerTools_RuntimeError, ox_list_user, ox_list_group, ox_call, listener_is_deactivated


ldap2ox = {
	'oxDisplayName': 'displayname',
	'title': 'title',
	'givenName': 'givenname',
	'oxMiddleName': 'middle_name',
	'sn': 'surname',
	'oxSuffix': 'suffix',
	'oxNickName': 'nickname',
	'o': 'company',
	'oxDepartment': 'department',
	'oxPosition': 'position',
	'oxProfession': 'profession',
	'oxBranches': 'branches',
	'oxMobileBusiness': 'cellular_telephone1',
	'mobile': 'cellular_telephone2',
	'oxTelephoneCompany': 'telephone_company',
	'oxTelephoneOther': 'telephone_other',
	'oxTelephoneCar': 'telephone_car',
	'oxFaxBusiness': 'fax_business',
	'oxFaxHome': 'fax_home',
	'oxFaxOther': 'fax_other',
	'oxManagerName': 'manager_name',
	'oxTelephoneAssistant': 'telephone_assistant',
	'mailPrimaryAddress': 'email1',
	'oxEmail2': 'email2',
	'oxEmail3': 'email3',
	'oxInstantMessenger1': 'instant_messenger1',
	'oxInstantMessenger2': 'instant_messenger2',
	'oxTelephoneIp': 'telephone_ip',
	'pager': 'telephone_pager',
	'oxTelephoneTelex': 'telephone_telex',
	'oxTelephoneTtydd': 'telephone_ttytdd',
	'oxUrl': 'url',
	'oxTaxId': 'tax_id',
	'oxCommercialRegister': 'commercial_register',
	'oxSalesVolume': 'sales_volume',
	'oxMarialStatus': 'marital_status',
	'oxSpouseName': 'spouse_name',
	'street': 'street_business',
	'postalCode': 'postal_code_business',
	'l': 'city_business',
	'oxStateBusiness': 'state_business',
	'oxCountryBusiness': 'country_business',
	'oxStreetHome': 'street_home',
	'oxPostalCodeHome': 'postal_code_home',
	'oxCityHome': 'city_home',
	'oxStateHome': 'state_home',
	'oxCountryHome': 'country_home',
	'oxStreetOther': 'street_other',
	'oxPostalCodeOther': 'postal_code_other',
	'oxCityOther': 'city_other',
	'oxStateOther': 'state_other',
	'oxCountryOther': 'country_other',
	'oxUserfield01': 'userfield01',
	'oxUserfield02': 'userfield02',
	'oxUserfield03': 'userfield03',
	'oxUserfield04': 'userfield04',
	'oxUserfield05': 'userfield05',
	'oxUserfield06': 'userfield06',
	'oxUserfield07': 'userfield07',
	'oxUserfield08': 'userfield08',
	'oxUserfield09': 'userfield09',
	'oxUserfield10': 'userfield10',
	'oxUserfield11': 'userfield11',
	'oxUserfield12': 'userfield12',
	'oxUserfield13': 'userfield13',
	'oxUserfield14': 'userfield14',
	'oxUserfield15': 'userfield15',
	'oxUserfield16': 'userfield16',
	'oxUserfield17': 'userfield17',
	'oxUserfield18': 'userfield18',
	'oxUserfield19': 'userfield19',
	'oxUserfield20': 'userfield20',
	'employeeType': 'categories',
	'oxNote': 'note',
	}

ldap2oxmulti = {
	'telephoneNumber': 'telephone_business',
	'homePhone': 'telephone_home',
}


name = 'ox-user'
description = 'User settings for OX'
filter = '(&(objectClass=inetOrgPerson)(objectClass=oxUserObject))'
attributes = ['uid', 'uidNumber', 'gidNumber', 'mailAlternativeAddress', 'roomNumber', 'employeeNumber',
	'univentionPolicyReference', 'isOxUser', 'oxBirthday', 'oxAnniversary', 'oxNumOfChildren',
	'oxAccess', 'oxAccessSyncMl', 'oxAccessResources', 'oxLanguage', 'oxTimeZone', 'oxAccessUSM', 'oxContextIDNum']
attributes.extend(ldap2ox.keys())
attributes.extend(ldap2oxmulti.keys())
# use the modrdn listener extension
modrdn = "1"

ldapserver = listener.configRegistry['ldap/server/name']
ldapbase = listener.configRegistry['ldap/base']
myhostname = listener.configRegistry['hostname']
mydomainname = listener.configRegistry['domainname']
myfqdn = '%s.%s' % (myhostname, mydomainname)
smtp_server = listener.configRegistry.get('ox/listener/smtp-server', 'localhost')
weblang = listener.configRegistry.get("locale/default", "").split(':')[0]

csv_user_row_uid = "Id"
csv_user_row_username = "Name"
csv_user_row_default_sender_address = "DefaultSenderAddress"
csv_group_row_groupname = "name"
csv_group_row_members = "members"


ACCESS_DEFINITIONS_FILE = '/opt/open-xchange/etc/ModuleAccessDefinitions.properties'
_access_definitions = dict(ts=None, defs=None)

logger = None


@ListenerQueue(name, stop_on_error=False)
def handler(dn, new, old, command):
	global logger

	if listener_is_deactivated(name):
		return

	logger = OXLogger(name)
	logger.qualifier = None

	# language
	os.environ["LANG"] = weblang

	oldIsOxUser = False
	if old:
		logger.info('Old present.')
		if old.get('isOxUser') and old['isOxUser'][0] == 'OK':
			oldIsOxUser = True

	newIsOxUser = False
	if new:
		logger.info('New present.')
		if new.get('isOxUser') and new['isOxUser'][0] == 'OK':
			newIsOxUser = True
	logger.process('New Object is OX-User: {}'.format(newIsOxUser))
	logger.process('Old Object is OX-User: {}'.format(oldIsOxUser))

	if not oldIsOxUser and not newIsOxUser:
		# Neither old nor new is OX user. Nothing to do.
		return

	blacklisted = ['Administrator', 'join-backup', 'join-slave', 'spam']

	username = ''
	if old.get('uid'):
		username = old['uid'][0]
		logger.process('old username: {}'.format(username))
	if new.get('uid'):
		username2 = new['uid'][0]
		logger.process('new username: {}'.format(username2))
		if username2 != username:
			if username2 in blacklisted or username in blacklisted:
				logger.warn('not allowed to change users in blacklist ({})'.format(username))
				return
		username = username2
	logger.qualifier = username

	groupwareoptions = []

	email = '@'
	if new and new.get('mailPrimaryAddress'):
			email = new['mailPrimaryAddress'][0]
			logger.process('new mailPrimaryAddress: {}'.format(email))
			groupwareoptions.append("--email=%s" % email)
			groupwareoptions.append("--imaplogin=%s" % email)

			aliases = '%s' % email
			aliaseslist = [email]
			if new.get('mailAlternativeAddress'):
				mailalternativeaddress = new['mailAlternativeAddress']
				for alias in mailalternativeaddress:
					aliases = '%s,%s' % (aliases, alias)
					aliaseslist.append(alias)
				logger.process('aliases {}'.format(aliases))
			groupwareoptions.append("--aliases=%s" % aliases)

	oldemail = '@'
	if old and old.get('mailPrimaryAddress'):
			oldemail = old['mailPrimaryAddress'][0]
			logger.process('old mailPrimaryAddress: {}'.format(oldemail))
	#
	# Context bestimmen:
	#

	contextnum = ''
	oxhomeserver = ''
	admuser = ''
	admpass = ''
	oldcontextnum = ''
	oldoxhomeserver = ''
	oldadmuser = ''
	oldadmpass = ''
	if new:
		domain = email.split('@')[1]
	else:
		domain = oldemail.split('@')[1]
	olddomain = oldemail.split('@')[1]

	with AsRoot():
		try:
			ldapCon = univention.uldap.getMachineConnection(ldap_master=False)
		except IOError as ex:
			raise ListenerTools_RuntimeError(ex)

	if new:
		contextnum = get_context_id_from_user(new)
		if not contextnum:
			logger.debug('No context in NEW, using domain "{}" to find context...'.format(domain))
			contextnum = get_context_id_from_domain(ldapCon, domain)
		oxhomeserver = get_ox_home_server_for_context(ldapCon, contextnum)
		admuser, admpass = get_context_credentials(contextnum)

	if old:
		oldcontextnum = get_context_id_from_user(old)
		if not oldcontextnum:
			logger.debug('No context in OLD, using domain "{}" to find context...'.format(olddomain))
			oldcontextnum = get_context_id_from_domain(ldapCon, olddomain)
		if oldcontextnum:
			oldoxhomeserver = get_ox_home_server_for_context(ldapCon, oldcontextnum)
			oldadmuser, oldadmpass = get_context_credentials(oldcontextnum)

	# Entscheidung:
	#    old new ##
	# 1)  -   1  ## E  kein altes Objekt, lokaler Kontext: new erzeugen
	# 2)  a   1  ## E  altes Objekt nicht lokal, neues Objekt lokal: new erzeugen
	# 3)  1   -  ## L  kein neues Objekt, lokaler Kontext: old loeschen
	# 4)  1   a  ## L  altes Objekt in lokalem Kontext, neues Objekt nicht in lokalem Kontext: old loeschen
	# 5)  1   2  ## LE alter und neuer Kontext lokal: old loeschen, new erzeugen
	# 6)  1   1  ## C  altes + neues Objekt, beide lokaler Kontext: change auf new
	# 7)  -   a  ## -  kein altes Objekt, nicht lokaler Kontext: nichts
	# 8)  a   -  ## -  altes Objekt geloescht, nicht lokaler Kontext: nichts
	# 9)  a   b  ## -  alter und neuer Kontext nicht lokal: nichts

	authArgs = [
		"--adminuser=%s" % admuser,
		"--adminpass=%s" % admpass,
		"--contextid=%s" % contextnum,
	]
	oldAuthArgs = [
		"--adminuser=%s" % oldadmuser,
		"--adminpass=%s" % oldadmpass,
		"--contextid=%s" % oldcontextnum,
	]

	logger.info('email={}'.format(email))
	logger.info('domain={}'.format(domain))
	logger.info('contextnum={}'.format(contextnum))
	logger.info('oxhomeserver={}'.format(oxhomeserver))
	logger.info('oldemail={}'.format(oldemail))
	logger.info('olddomain={}'.format(olddomain))
	logger.info('oldcontextnum={}'.format(oldcontextnum))
	logger.info('oldoxhomeserver={}'.format(oldoxhomeserver))
	logger.info('myfqdn={}'.format(myfqdn))
	logger.info('admuser={}'.format(admuser))
	logger.info('oldadmuser={}'.format(oldadmuser))

	# old lokal? (3) (4) (5) (6)
	if myfqdn == oldoxhomeserver:
		logger.info('old local (3,4,5,6)')
		if contextnum != oldcontextnum or not newIsOxUser:
			# loeschen (3) (4) (5)
			logger.info('loeschen (3) (4) (5)')

			oldoxuidnumber = ''
			for row in ox_list_user(oldadmuser, oldadmpass, oldcontextnum, searchpattern=username, ignorecase=True):
				if row[csv_user_row_username] == username:
					oldoxuidnumber = row[csv_user_row_uid]
			logger.process('Processing uid {} for old object'.format(oldoxuidnumber))

			for row in ox_list_group(oldadmuser, oldadmpass, oldcontextnum):
				# logger.info('Processing row: {}'.format(row))
				members = row[csv_group_row_members].split(',')
				group = row[csv_group_row_groupname]
				if oldoxuidnumber in members:
					if list.__len__(members) <= 1:
						logger.process('removing ox group "{}" from context {}'.format(group, oldcontextnum))
						groupdelete = [
							"deletegroup",
							"--name=%s" % group,
						] + oldAuthArgs
						ox_call(groupdelete, "com.openexchange.admin.rmi.exceptions.NoSuchObjectException")

			if username == oldadmuser:
				logger.info('Not deleting admin-user {}'.format(username))
			else:
				if command == 'r':
					# object has moved, write ldap uidNumber into filename and oxuidnumber in this file.
					logger.info('Command = r')
					lockfilename = '/var/lock/deleteuid-%s' % old['uidNumber'][0]
					with AsRoot():
						with open(lockfilename, 'w') as lockfilehandle:
							lockfilehandle.write(oldoxuidnumber)
					logger.process('setting lock file for user uidNumber: {}'.format(old['uidNumber'][0]))
				else:
					deleteuser = [
						"deleteuser",
						"--username=%s" % username,
					] + oldAuthArgs
					logger.process('deleting ox user "{}" from context {}'.format(username, oldcontextnum))
					ox_call(deleteuser, "com.openexchange.admin.rmi.exceptions.NoSuchObjectException")

	# new lokal? (1) (2) (5) (6)
	if myfqdn == oxhomeserver:
		logger.info('new lokal (1) (2) (5) (6)')
		# ldapsearch -x gidNumber=5001
		# ldapsearch -x "(&(objectClass=posixGroup)(gidNumber=5001))" cn
		# ldapsearch -x "(&(objectClass=posixGroup)(memberUid=aaa))" cn

		if new.get('gidNumber'):
			gidnumber = new['gidNumber'][0]
		else:
			gidnumber = old['gidNumber'][0]

		filter_s = filter_format('(&(gidNumber=%s)(objectClass=posixGroup))', (str(gidnumber),))
		result_id = ldapCon.search(base=ldapbase, filter=filter_s, attr=['cn'])
		if result_id:
			prigroup = result_id[0][1].values()[0][0]
			# logger.info('prigroup={}'.format(prigroup))
		filter_s = filter_format('(&(memberUid=%s)(objectClass=posixGroup))', (username,))
		result_id = ldapCon.search(base=ldapbase, filter=filter_s, attr=['cn'])
		if result_id:
			othergroups = result_id
			# logger.info('othergroups={}'.format(othergroups))

		imapserver = 'localhost'
		if listener.configRegistry.get('ox/role') != 'oxae':
			filter_s = filter_format('(&(cn=%s)(objectClass=oxMailDomainServerSettings))', (domain,))
			result_id = ldapCon.search(base='cn=domain,cn=mail,%s' % ldapbase, filter=filter_s, attr=['oxImapServer'])
			if result_id:
				imapserver = result_id[0][1].values()[0][0]
			logger.info('imapserver={}'.format(imapserver))

		groupwareoptions.append("--imapserver=%s" % imapserver)
		groupwareoptions.append("--smtpserver=%s" % smtp_server)

		if not listener.configRegistry.is_true("mail/dovecot"):
			for reference in new.get('univentionPolicyReference', []):
				try:
					filter_s = '(objectClass=univentionMailQuota)'
					result_id = ldapCon.search(base=reference, filter=filter_s, attr=['univentionMailQuotaMB'])
					if result_id[0][1].get('univentionMailQuotaMB', [''])[0]:
						logger.info('Calling USER={} /usr/sbin/univention-cyrus-set-quota'.format(username))
						listener.setuid(0)
						oldUSER = os.environ["USER"]
						os.environ["USER"] = username
						subprocess.call('/usr/sbin/univention-cyrus-set-quota &', shell=True)
						os.environ["USER"] = oldUSER
						listener.unsetuid()
				except:
					pass

		if new.get('oxBirthday'):
			if new['oxBirthday'][0]:
				birthday = new['oxBirthday'][0]
				text = birthday
				if '-' in text:
					birthdayiso = birthday
				else:
					day, month, year = map(int, text.split('.'))
					birthdayiso = '%s-%s-%s' % (year, month, day)
				groupwareoptions.append("--birthday=%s" % birthdayiso)

		if new.get('oxAnniversary'):
			if new['oxAnniversary'][0]:
				birthday = new['oxAnniversary'][0]
				text = birthday
				if '-' in text:
					birthdayiso = birthday
				else:
					day, month, year = map(int, text.split('.'))
					birthdayiso = '%s-%s-%s' % (year, month, day)
				groupwareoptions.append("--anniversary=%s" % birthdayiso)

		if new.get('oxNumOfChildren'):
			groupwareoptions.append("--number_of_children=%s" % new['oxNumOfChildren'][0])
		elif old and old.get('oxNumOfChildren'):
			groupwareoptions.append("--number_of_children=")

		if new.get('roomNumber'):
			groupwareoptions.append("--room_number=%s" % new['roomNumber'][0])
		elif old and old.get('roomNumber'):
			groupwareoptions.append("--room_number=")

		if new.get('employeeNumber'):
			groupwareoptions.append("--number_of_employee=%s" % new['employeeNumber'][0])
		elif old and old.get('employeeNumber'):
			groupwareoptions.append("--number_of_employee=")

		oxuidnumber = get_ox_id(
			admuser or oldadmuser,
			admpass or oldadmpass,
			contextnum or oldcontextnum,
			username)
		if oxuidnumber:
			logger.info('old object present (oxuidnumber={})'.format(oxuidnumber))

		if new.get('oxAccess'):
			oxaccess = new['oxAccess'][0]
			oxaccess_usm = new.get('oxAccessUSM', ['Not'])[0] == 'OK'
			groupwareoptions.extend(get_access_cmdline_args(oxaccess, oxuidnumber, oxaccess_usm))
		else:
			groupwareoptions.append("--mailenabled=false")

		if new.get('oxAccessSyncMl'):
			groupwareoptions.append("--access-syncml=on")
		else:
			groupwareoptions.append("--access-syncml=off")

		if new.get('oxAccessResources'):
			groupwareoptions.append("--access-edit-resource=on")
		else:
			groupwareoptions.append("--access-edit-resource=off")

		for ldapattr, oxkey in ldap2ox.items():
			if new and new.get(ldapattr):
				groupwareoptions.append("--%s=%s" % (oxkey, new.get(ldapattr)[0]))
			elif old and old.get(ldapattr):
				groupwareoptions.append("--%s=" % oxkey)

		# if more than one value is present then set the first value into "${key}1" and the second into "${key}2"
		for ldapattr, oxmultikey in ldap2oxmulti.items():
			if new and new.get(ldapattr):
				groupwareoptions.append("--%s1=%s" % (oxmultikey, new.get(ldapattr)[0]))
				if len(new.get(ldapattr)) > 1:
					groupwareoptions.append("--%s2=%s" % (oxmultikey, new.get(ldapattr)[1]))
			elif old and old.get(ldapattr):
				groupwareoptions.append("--%s1=" % oxmultikey)
				groupwareoptions.append("--%s2=" % oxmultikey)

		groupwareoptions.append("--gui_spam_filter_capabilities_enabled=true")

		# Test for lockfile if userobject has moved.
		userhasmoved = False
		if new:
			try:
				lockfilename = '/var/lock/deleteuid-%s' % new['uidNumber'][0]
				logger.process('lockfilename: {}'.format(lockfilename))
				lockfiletext = open(lockfilename).read()
				logger.process('lockfiletext: {}'.format(lockfiletext))
				if lockfiletext:
					oxuidnumber = lockfiletext
					userhasmoved = True
					with AsRoot():
						os.remove(lockfilename)
			except:
				pass
		# newcontext != oldcontext ?
		# test, ob benutzer im context schon existiert
		if userhasmoved:
			changeuser = [
				"changeuser",
				"--userid=%s" % oxuidnumber,
				"--username=%s" % username,
			] + authArgs
			logger.info('Modifying OX username {}, oxuid {} in context {}'.format(username, oxuidnumber, contextnum))
			ox_call(changeuser)

		if ((contextnum == oldcontextnum) and oxuidnumber) or userhasmoved and oldIsOxUser and newIsOxUser:
			# aendern (6)
			logger.info('aendern (6)')
			if new.get('mailPrimaryAddress'):
				defaultsenderaddress = new['mailPrimaryAddress'][0]
				# Aktuelle default-address auslesen

				oxdefaultsenderaddress = ''
				for row in ox_list_user(admuser, admpass, contextnum, searchpattern=username, ignorecase=True):
					if row[csv_user_row_username] == username:
						oxdefaultsenderaddress = row[csv_user_row_default_sender_address]

				if oxdefaultsenderaddress not in aliaseslist:
					groupwareoptions.append("--defaultsenderaddress=%s" % defaultsenderaddress)

			if new.get('oxLanguage'):
				language = new['oxLanguage'][0]
				oldlanguage = ''
				if old.get('oxLanguage'):
					oldlanguage = old['oxLanguage'][0]
				if language != oldlanguage:
					groupwareoptions.append("--language=%s" % language)

			if new.get('oxTimeZone'):
				timezone = new['oxTimeZone'][0]
				oldtimezone = ''
				if old.get('oxTimeZone'):
					oldtimezone = old['oxTimeZone'][0]
				if timezone != oldtimezone:
					groupwareoptions.append("--timezone=%s" % timezone)

			if email == oldemail:
				# changeuser returns error message if --email is given as argument but value dit not change
				# ===>
				# user user2 in context ?? could not be changed:
				# Server response:
				# 	 Primary mail address already exists in this context.
				newgroupwareoptions = []
				for i in groupwareoptions:
					if not i.startswith("--email="):
						newgroupwareoptions.append(i)
				groupwareoptions = newgroupwareoptions
			command = [
				"changeuser",
				"--username=%s" % username,
			] + authArgs + groupwareoptions
			logger.info('Modifying OX settings for uid "{}" in context {}'.format(username, contextnum))
			ox_call(command)
		else:
			# erzeugen (1) (2) (5)
			logger.info('erzeugen (1) (2) (5)')
			if newIsOxUser:

				if not oxuidnumber:
					command = ["createuser"] + authArgs
					logger.process('Creating ox user "{}" in context {}'.format(username, contextnum))
				else:
					command = ["changeuser"] + authArgs
					logger.info('Modifying OX settings for uid "{}" in context {}'.format(username, contextnum))

				if new.get('mailPrimaryAddress'):
					groupwareoptions.append("--defaultsenderaddress=%s" % new['mailPrimaryAddress'][0])

				if new.get('oxLanguage'):
					groupwareoptions.append("--language=%s" % new['oxLanguage'][0])

				if new.get('oxTimeZone'):
					groupwareoptions.append("--timezone=%s" % new['oxTimeZone'][0])

				if username == admuser:
					password = admpass
				else:
					password = "dummy"

				command.append("--username=%s" % username)
				command.append("--password=%s" % password)
				command = command + groupwareoptions

				ox_call(command)
				if not oxuidnumber:
					oxuidnumber = get_ox_id(admuser, admpass, contextnum, username)
				if not oxuidnumber:
					# this should never happen
					logger.error(
						'DB didn\'t return oxuidnumber for username "{username}" in context {context}. Probably '
						'not an OX user.\nnew: {new}\nold: {old}'.format(
							username=username, new=new, old=old, context=contextnum))
					raise Exception("User '{0}' doesn't exist after creation in context {1}.".format(username, contextnum))

		filter_s = filter_format('(&(memberUid=%s)(objectClass=posixGroup))', (username,))
		result_id = ldapCon.search(base=ldapbase, filter=filter_s, attr=['cn'])
		if result_id:
			# logger.info('groups found')
			oxgroups = [prigroup]
			for w in result_id:
				if w[1].values()[0][0] != prigroup:
					oxgroups.append(w[1].values()[0][0])
			logger.process('oxgroups: {}'.format(oxgroups))

			# Benutzer zu Gruppen hinzufügen, bzw. Gruppe mit Benutzer anlegen:
			if oxuidnumber:
				for group in oxgroups:
					# logger.info('processing group {}'.format(w[1].values()[0][0]))
					# Gruppe suchen:
					foundgroup = False
					for row in ox_list_group(admuser, admpass, contextnum):
						if row[csv_group_row_groupname].lower() == group.lower():
							foundgroup = True
							break

					# Displayname zu Gruppe finden:
					# 'description'
					# 'cn'
					filter_s = filter_format('(&(cn=%s)(objectClass=oxGroup))', (group,))
					result_id = ldapCon.search(base=ldapbase, filter=filter_s, attr=['cn', 'description'])
					# logger.info('result_id (group): {}({})'.format(result_id, group))
					groupdisplayname = group
					if result_id:
						if not foundgroup:
							logger.process('creating ox group: "{}" in context {}'.format(group, contextnum))
						if result_id[0][1].get('description'):
							groupdisplayname = result_id[0][1]['description'][0]
							# logger.info('groupsdisplayname: {}'.format(groupdisplayname))
						# gruppe existiert nicht und sicherheitshalber, falls die Gruppe doch bereits exisitiert.
						command = [
							"changegroup" if foundgroup else "creategroup",
							"--name=%s" % group,
							"--displayname=%s" % groupdisplayname,
							"--addmembers=%s" % oxuidnumber,
						] + authArgs
						logger.info('Handling group membership for uid "{}" in context {}'.format(username, contextnum))
						ox_call(command, "Member already exists in group")
			# Benutzer aus Gruppen entfernen:
			# OX-Gruppen des User-Objektes finden, in denen der Benutzer ist (oxgroups)
			# alle anderen Gruppen auf OX suchen (oxgroupmap)
			nooxgroups = []
			for row in ox_list_group(admuser, admpass, contextnum):
				if not row[csv_group_row_groupname] in oxgroups:
					if row[csv_group_row_groupname] != 'users':
						tmpoxgroup = [
							row[csv_group_row_groupname],
							row[csv_group_row_members],
						]
						nooxgroups.append(tmpoxgroup)
			logger.info('User in OX-groups: {}'.format(oxgroups))
			logger.info('User not in OX-groups: {}'.format(nooxgroups))
			# in diesen Gruppen auf OX suchen, ob Benutzer drin ist:
			for grouprow in nooxgroups:
				groupuserlist = grouprow[1].split(',')
				if oxuidnumber in groupuserlist:
					# ja: Benutzer aus Gruppe entfernen

					# Prüfen, ob Benutzer der letzte der Gruppe war:
					if list.__len__(groupuserlist) <= 1:
						# ja: Gruppe löschen
						logger.process('removing ox group: "{}" from context {}'.format(grouprow[0], contextnum))
						deletegroup = [
							"deletegroup",
							"--name=%s" % grouprow[0],
						] + authArgs
						ox_call(deletegroup, "com.openexchange.admin.rmi.exceptions.NoSuchObjectException")

					else:
						logger.info('removing ox user from group: "{}" from context {}'.format(grouprow[0], contextnum))
						changegroup = [
							"changegroup",
							"--name=%s" % grouprow[0],
							"--displayname=%s" % groupdisplayname,
							"--removemembers=%s" % oxuidnumber,
						] + authArgs
						ox_call(changegroup)


def get_context_credentials(contextid):
	admuser = 'oxadmin-context%s' % (contextid,)
	if admuser == 'oxadmin-context10':
		admuser = 'oxadmin'
	admpass = get_credentials('/etc/ox-secrets/context%s.secret' % (contextid,))
	return admuser, admpass


def get_ox_id(admuser, admpass, contextnum, username):
	for row in ox_list_user(admuser, admpass, contextnum, searchpattern=username, ignorecase=True):
		if row[csv_user_row_username] == username:
			return row[csv_user_row_uid]


def get_access_definitions():
	"""
	Parse ACCESS_DEFINITIONS_FILE for ox access definitions.

	:return: dict: access definition (str) -> capabilities to enable (list)
	"""
	ts = os.stat(ACCESS_DEFINITIONS_FILE).st_mtime  # will raise OSError if file inaccessible
	if not _access_definitions['ts'] or _access_definitions['ts'] < ts:
		_access_definitions['ts'] = ts
		_access_definitions['defs'] = dict(__all=set())
		regex = re.compile(r'^(\w+)=(.+)')
		for line in open(ACCESS_DEFINITIONS_FILE, 'rb'):
			m = regex.match(line)
			if not m:
				continue
			name = m.groups()[0].strip()
			capabilities = {x.strip() for x in m.groups()[1].split(',')}
			_access_definitions['defs'][name] = capabilities
			_access_definitions['defs']['__all'].update(capabilities)
		logger.info('found the following access definitions in "{}":\n{}'.format(
			ACCESS_DEFINITIONS_FILE, _access_definitions))
	return _access_definitions['defs']


def get_access_cmdline_args(access_combination_name, changeuser, mobility):
	"""
	Create cmdline arguments for changeuser regarding feature access.

	:param access_combination_name: str: UDM oxAccess, currently one of 'none',
	'webmail', 'pim', 'groupware', 'premium'.
	:param changeuser: bool: whether to create arguments for changeuser (True)
	or createuser (False)
	:param mobility: bool: whether to enable mobility access (UDM oxAccessUSM)
	:return: list: list of strings
	"""
	access_definitions = get_access_definitions()
	if access_combination_name == 'none':
		res = ['--mailenabled=false']
		disable_caps = access_definitions['__all']
	else:
		access_definition_name = 'oxseforucs_{}{}{}'.format(
			access_combination_name,
			'_plus' if access_combination_name == 'pim' else '',
			'_usm' if mobility else '')
		res = ['--mailenabled=true', '--access-combination-name', access_definition_name]
		try:
			disable_caps = access_definitions['__all'] - access_definitions[access_definition_name]
		except KeyError:
			raise ValueError('Unknown access definition %r.', access_definition_name)
	if changeuser:
		res.extend(['--capabilities-to-drop', ','.join(disable_caps)])
	if not mobility and 'usm' in disable_caps:
		res.extend(['--access-usm', 'off'])
	return res


def get_default_context_id():
	return listener.configRegistry.get('ox/context/id', '10')


def get_context_id_from_domain(ldapCon, domainname):
	filter_s = filter_format('(&(objectClass=oxMailDomainServerSettings)(cn=%s))', (domainname,))
	query_result = ldapCon.search(base='cn=domain,cn=mail,%s' % ldapbase, filter=filter_s, attr=['oxContextName'])
	if not query_result:
		return ''
	context_name = query_result[0][1].values()[0][0]

	filter_s = filter_format('(&(objectClass=oxContext)(cn=%s))', (context_name,))
	query_result = ldapCon.search(base='cn=open-xchange,%s' % ldapbase, filter=filter_s, attr=['oxContextIDNum'])
	if not query_result:
		return ''
	return query_result[0][1].values()[0][0]


def get_context_id_from_user(old_or_new):
	return old_or_new.get('oxContextIDNum', [''])[0]


def get_ox_home_server_for_context(ldapCon, context_id):
	result = ldapCon.search(
		base='cn=open-xchange,%s' % ldapbase,
		filter=filter_format('(&(objectClass=oxContext)(oxContextIDNum=%s))', (str(context_id),)),
		attr=['oxHomeServer'])
	if result:
		return result[0][1]['oxHomeServer'][0]
	else:
		raise RuntimeError('Cannot find context information for ID %r.' % (context_id,))


if __name__ == '__main__':
	handler.process_pending()
